#!/usr/bin/env python

__author__ = "nosmo@nosmo.me"

import iptc
import hostsfile

import banobjects

import time
import logging
import threading
import traceback

# Clean rules that have expired

# in minutes
BANTIME = 2

BANLIMIT = 10

class BanCleaner(object):

    def _iptc_clean_bans(self, interface=None):

        banlist = []

        with self.iptables_lock:
            table = iptc.Table(iptc.Table.FILTER, autocommit=False)
            chain = iptc.Chain(table, "INPUT")
            rules = chain.rules
            for index, rule in enumerate(rules):
                # This does two selects
                # dumb but fix later.

                now = int(time.time())
                ban = self.ban_object(rule.src.split("/")[0])
                if not ban.banstart:
                    continue

                if (now - ban.banstart) > self.timelimit:
                    logging.info("Unbanning %s as the ban has expired", ban.ipaddress)
                    banlist.append(ban)
                    logging.debug("Unbanned %s", ban.ipaddress)
                if index > BANLIMIT:
                    # Rate limit a little
                    break

            for ban in banlist:
                ban.unban()
            table.commit()

        return True

    def _hosts_clean_bans(self, interface=None):

        hostsban = hostsfile.HostsDeny()
        for banentry in hostsban:
            ban = self.ban_object(banentry[1])
            if not ban.banstart:
                continue

            now = int(time.time())
            if (now - ban.banstart) > self.timelimit:
                logging.info("Unbanning %s as the ban has expired", ban.ipaddress)
                ban.unban()

    def _iptables_cmd_clean_bans(self, interface=None):
        for rule in banobjects.IPTablesCommandBanEntry.list(self.timelimit):
            ruletodelete = banobjects.IPTablesCommandBanEntry(rule)
            logging.info("Unbanning %s as the ban has expired", ruletodelete.ipaddress)
            ruletodelete.unban(interface)

    #TODO make lock optional
    def __init__(self, bantime, backend, lock, interface):
        self.bantime = bantime
        self.interface = interface
        self.ban_object = banobjects.entries[backend]
        self.timelimit = bantime

        self.iptables_lock = lock

        self.clean_bans = {
            "hostsfile": self._hosts_clean_bans,
            "iptables": self._iptc_clean_bans,
            "iptables_cmd": self._iptables_cmd_clean_bans
            }[backend]

def main():

    #mainlogger = logging.getLogger()
    #logging.basicConfig(level=logging.DEBUG)
    #ch = logging.StreamHandler(sys.stdout)
    #ch.setLevel(logging.DEBUG)
    #formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    #ch.setFormatter(formatter)
    #mainlogger.addHandler(ch)

    b = BanCleaner(BANTIME, "iptables", threading.Lock(), "eth+")
    while True:
        b.clean_bans()
        time.sleep(60)

if __name__ == "__main__":
    main()



#!/usr/bin/env python

__author__ = "nosmo@nosmo.me"

import zmq
import ipaddr
from zmq.eventloop import ioloop, zmqstream

import datetime
import logging
import re
import sys
import threading

import banobjects

BIND_STRINGS = ["tcp://127.0.0.1:22620"]

class BanFetcher(threading.Thread):

    #TODO make lock optional
    def __init__(self, bindstrings,
                 interface, backend,
                 whitelist, lock):
        self.bindstrings = bindstrings
        self.interface = interface
        self.backend = backend
        self.ban_object = banobjects.entries[backend]

        self.whitelist = {"networks": [], "addresses": []}
        for entry in whitelist:
            if "/" in entry:
                # This check sucks but it'll do for now. ipaddr will
                # throw a suitable exception if this doesn't work
                self.whitelist["networks"].append(ipaddr.IPNetwork(entry))
            else:
                self.whitelist["addresses"].append(ipaddr.IPAddress(entry))

        self.sockets = {}

        for bindstring in bindstrings:

            context = zmq.Context()
            self.sockets[bindstring] = context.socket(zmq.SUB)
            subscriber = zmqstream.ZMQStream(self.sockets[bindstring])

            if hasattr(zmq, "RCVHWM"):
                self.sockets[bindstring].setsockopt(zmq.RCVHWM, 2000)
            if hasattr(zmq, "SNDHWM"):
                self.sockets[bindstring].setsockopt(zmq.SNDHWM, 2000)
            if hasattr(zmq, "HWM"):
                self.sockets[bindstring].setsockopt(zmq.HWM, 2000)

            # SWAP is removed in zmq :(
            if hasattr(zmq, "SWAP"):
                self.sockets[bindstring].setsockopt(zmq.SWAP, 200*2**10)

            self.sockets[bindstring].setsockopt(zmq.SUBSCRIBE, "swabber_bans")
            self.sockets[bindstring].connect(bindstring)
            subscriber.on_recv(self.subscription)

        self.iptables_lock = lock

        threading.Thread.__init__(self)
        self.daemon = True

    def subscription(self, message):
        if len(message) != 2:
            logging.debug("ZMQ received invalid message: %s", message)
            return False

        action, ipaddress = message

        ipaddress = ipaddress.strip()
        ipmatch = re.compile("^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$")
        if not ipmatch.match(ipaddress):
            logging.error("Failed to validate IP address %s - rejecting", ipaddress)
            return False

        ipaddr_obj = ipaddr.IPAddress(ipaddress)
        for network in self.whitelist["networks"]:
            if ipaddr_obj in network:
                logging.info("Not banning IP %s because it is in whitelisted network %s",
                             ipaddress, str(network))
                return False
        for whitelist_ip in self.whitelist["addresses"]:
            if ipaddr_obj == whitelist_ip:
                logging.info("Not banning IP %s because it is whitelisted", ipaddress)
                return False

        if action == "swabber_bans":
            logging.debug("Received ban for %s", message[1])
            thenow = datetime.datetime.now()

            with self.iptables_lock:
                ban = self.ban_object(ipaddress)
                logging.debug("Created banentry for %s", ipaddress)

                logging.debug("Fetcher got iptables lock")
                try:
                    if ban.new_ban:
                        logging.info("Created ban for %s", ipaddress)
                        ban.ban(self.interface)
                    else:
                        oldstart = ban.banstart
                        ban.unban(self.interface)
                        ban.ban(self.interface)
                        newstart = ban.banstart
                        logging.info("Extended ban for %s from %d to %d", ipaddress,
                                     oldstart, newstart)

                except self.ban_object.fault_exception as e:
                    logging.error("Failed to initialise ban - do we lack permissions?: %s", e)

        else:
            logging.error("Got an invalid message header: %s", message)
            return False
        return True

    def stop_running(self):
        self.loop.stop()

    def run(self):
        self.loop = ioloop.IOLoop.instance().start()

def main():
    verbose = False

    if verbose:
        mainlogger = logging.getLogger()

        logging.basicConfig(level=logging.DEBUG)
        ch = logging.StreamHandler(sys.stdout)
        ch.setLevel(logging.DEBUG)
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        ch.setFormatter(formatter)
        mainlogger.addHandler(ch)

    bfetcher = BanFetcher(BIND_STRINGS[0], "eth+", "iptables", threading.Lock())
    bfetcher.run()

if __name__ == "__main__":

    main()



#!/usr/bin/env python 

import unittest
import commands
import threading
import os
import time

from banobjects import IPTablesCommandBanEntry
from bancleaner import BanCleaner
from banfetcher import BanFetcher

BAN_IP = "10.123.45.67"
BINDSTRING = ["tcp://127.0.0.1:22620"]
INTERFACE = "eth+"
BAN_BACKEND = "iptables_cmd"

class IPTablesCommandBanTests(unittest.TestCase):

    ''' Tests for the IPTables command-based ban objects.
    '''

    def test_ban(self):
        ''' Create a ban and remove it. '''

        ban = IPTablesCommandBanEntry(BAN_IP)
        ban.ban(interface=INTERFACE)

        status, output = commands.getstatusoutput("/sbin/iptables -L -n")
        self.assertEqual(status, 0, 
                         msg="Failed to run iptables command: %s" % output)
        self.assertIn(BAN_IP, output, msg="IP address not banned")
        ban.unban(interface=INTERFACE)
        status, output = commands.getstatusoutput("/sbin/iptables -L -n")
        self.assertNotIn(BAN_IP, output, msg="IP address was not unbanned")

    def test_whitelist_ip(self):
        ''' Test the whitelisting of a single IP ''' 
        
        lock = threading.Lock()
        ban_fetcher = BanFetcher(BINDSTRING, INTERFACE,
                                 BAN_BACKEND, ["10.0.220.1", "10.0.222.1/24"],
                                 lock)
        # An address that shouldn't be banned
        self.assertFalse(
            ban_fetcher.subscription(("swabber_bans", "10.0.220.1")), 
            msg="IP that should have been whitelisted was banned!")

    def test_whitelist_network(self): 
        ''' Test the whitelisting of a whole network '''

        lock = threading.Lock()
        ban_fetcher = BanFetcher(BINDSTRING, INTERFACE,
                                 BAN_BACKEND, ["10.0.220.1", "10.0.222.1/24"],
                                 lock)
        # An address in a network that shouldn't be banned
        self.assertFalse(
            ban_fetcher.subscription(("swabber_bans", "10.0.222.3")), 
            msg="IP in a network that should have been whitelisted was banned!")

    def test_whitelist_notwhitelisted(self): 
        ''' Test the unwhitelisting of an IP to ensure it isn't whitelisted '''

        lock = threading.Lock()
        ban_fetcher = BanFetcher(BINDSTRING, INTERFACE,
                                 BAN_BACKEND, ["10.0.220.1", "10.0.222.1/24"],
                                 lock)

        # An address that should be banned
        self.assertTrue(
            ban_fetcher.subscription(("swabber_bans", "10.0.220.2")), 
            msg="Unwhitelisted IP was whitelisted")

        # clean up after ourselves
        ban = IPTablesCommandBanEntry("10.0.220.2")
        ban.unban(interface=INTERFACE)

class CleanTests(unittest.TestCase):

    ''' Test the bancleaner object '''

    def test_clean(self):
        ''' Test the cleaning of bans after a very short time window '''

        ban_len = 1
        ban = IPTablesCommandBanEntry(BAN_IP)
        ban.ban(INTERFACE)
        time.sleep(ban_len*2)

        cleaner = BanCleaner(ban_len, BAN_BACKEND, threading.Lock(), INTERFACE)
        cleaner.clean_bans(INTERFACE)

        status, output = commands.getstatusoutput("/sbin/iptables -L -n")
        self.assertEqual(status, 0, 
                         msg="Failed to run iptables command: %s" % output)
        self.assertNotIn(BAN_IP, output, msg="Ban was not reset by cleaner")

def main():
    ''' Run some tests boyo '''

    if os.getuid() != 0:
        print "Tests must be run as root"
        raise SystemExit
    else:
        unittest.main()

if __name__ == '__main__':
    main()



#!/usr/bin/env python

__author__ = "nosmo@nosmo.me"

import iptc
import time
import commands
import logging
import hostsfile

def get_iptables_version():
    status, output = commands.getstatusoutput('/sbin/iptables --version')
    if status != 0:
        return None
    version = output.strip().split(" ")[1]
    if version.startswith("v"):
        version = version[1:]
    return [ int(i) for i in version.split(".") ]

IPTABLES_VERSION = get_iptables_version()

def iptables_has_wait():
    introduced_in = [1, 4, 20]
    if not IPTABLES_VERSION:
        return False
    if IPTABLES_VERSION[0] > introduced_in[0]:
        return True
    elif IPTABLES_VERSION[1] > introduced_in[1]:
        return True
    elif IPTABLES_VERSION[2] >= introduced_in[2]:
        return True
    else:
        return False

class IPTablesCommandBanEntry(object):

    fault_exception = Exception

    def __init__(self, ipaddress):
        self.ipaddress = ipaddress
        self.banstart = None

        self.new_ban = True

        for rule, start in self.list().iteritems():
            if rule == self.ipaddress:
                self.banstart = int(start)
                self.new_ban = False

    @staticmethod
    def list(timelimit=None, wait=True):
        # timelimit - limit the listing to entries in the list
        # $timelimit seconds

        # wait - use the iptables wait to get a lock for listing. Will
        # slow down some operations, but means they will definitely
        # happen. Has no effect if iptables doesn't support it.

        iptables_command = "/sbin/iptables -L -n"
        if wait and iptables_has_wait():
            iptables_command += " -w"

        rulesdict = {}
        status, output = commands.getstatusoutput(iptables_command)
        if status:
            # This usually indicates that there's an xtables lock
            logging.error("Couldn't list iptables rules! %s", output)
            return rulesdict

        droprules = [ i for i in output.split("\n") if i.startswith("DROP") and "swabber" in i ]
        for rule in droprules:
            action, proto, opt, src, dest, _, swabber, _ = rule.split()
            if ":" not in swabber:
                raise Exception("Malformed swabber rule in iptables! %s" % swabber)
            _, start = swabber.split(":")
            start = int(start.strip())
            if not timelimit or (timelimit and (time.time() - start > timelimit)):
                rulesdict[src] = start

        return rulesdict

    def ban(self, interface=None, wait=True):
        interface_section = "-i %s" % interface if interface else ""

        now = int(time.time())
        iptables_command = ("iptables -I INPUT -s %s %s -j DROP -m comment"
                   " --comment \"swabber:%d\"") % (
                       self.ipaddress, interface_section, now)

        if wait and iptables_has_wait():
            iptables_command += " -w"

        status, output = commands.getstatusoutput(iptables_command)
        self.banstart = now

        if status:
            raise Exception("Couldn't set iptables rule for %s (command %s): %s" % (
                self.ipaddress, command, output))
        return True

    def unban(self, interface=None, wait=True):
        interface_section = "-i %s" % interface if interface else ""

        iptables_command = "iptables -D INPUT -s %s -j DROP -m comment --comment \"swabber:%d\" %s" % (self.ipaddress, self.banstart, interface_section)
        if wait and iptables_has_wait():
            iptables_command += " -w"

        status, output = commands.getstatusoutput(iptables_command)
        if status:
            raise Exception("failed to unban IP %s: %s command %s" % (self.ipaddress, output, iptables_command))
        return True

    def __repr__(self):
        return "<BanEntry('%s', %s)>" % (self.ipaddress,
                                         self.banstart)

class HostsBanEntry(object):

    fault_exception = IOError

    def __init__(self, ipaddress):
        self.hostsfile = hostsfile.HostsDeny()
        self.ipaddress = ipaddress
        self.banstart = None
        self.new_ban = True
        if ipaddress in self.hostsfile:
            hostsfileentry = self.hostsfile[ipaddress]
            if hostsfileentry[2] and "swabber" in hostsfileentry[2]:
                self.banstart = int(hostsfileentry[2].split(":")[1])
                self.new_ban = False

    def ban(self, interface=None):
        #interface is a dummy
        self.banstart = int(time.time())
        comment = "swabber:%s" % self.banstart
        self.hostsfile.add(self.ipaddress, comment=comment)

        return True

    def unban(self):
        self.hostsfile -= self.ipaddress

    def __repr__(self):
        return "<BanEntry('%s', %s)>" % (self.ipaddress,
                                         self.banstart)

class IPTCBanEntry(object):

    fault_exception = iptc.IPTCError

    def __init__(self, ipaddress):
        self.ipaddress = ipaddress

        table = iptc.Table(iptc.Table.FILTER, autocommit=False)
        chain = iptc.Chain(table, "INPUT")
        self.rule = None
        self.banstart = None
        self.new_ban = True

        rules = chain.rules
        for rule in rules:
            if rule.src == "%s/255.255.255.255" % self.ipaddress:
                if rule.matches:
                    comment = rule.matches[0].comment
                    #TODO regexp matching
                    if "swabber:" in comment:
                        self.chain = chain
                        self.rule = rule
                        self.banstart = int(comment.split(":")[1].strip('"'))
                        self.new_ban = False

        #if not self.rule and not self.banstart:
        #    # We're a new rule
        #    self.banstart = int(time.time())

        if not self.rule:
            self.table = None
            self.chain = None

    #TODO
    #@staticmethod
    #def _static_ban(ban, interface):

    def ban(self, interface):
        self.banstart = int(time.time())

        self.table = iptc.Table(iptc.Table.FILTER, autocommit=False)
        self.chain = iptc.Chain(table, "INPUT")
        rule = iptc.Rule()
        rule.in_interface = interface
        rule.src = self.ipaddress
        target = iptc.Target(rule, "DROP")
        rule.target = target

        rulecomment = rule.create_match("comment")
        rulecomment.comment = "swabber:%s" % self.banstart

        self.rule = rule

        self.chain.insert_rule(rule)
        self.table.commit()

        return True

    def unban(self):
        if not self.rule:
            return False
        self.chain.delete_rule(self.rule)
        return True

    def __repr__(self):
        return "<BanEntry('%s', %s)>" % (self.ipaddress,
                                         self.banstart)

entries = {
    "iptables_cmd": IPTablesCommandBanEntry,
    "iptables": IPTCBanEntry,
    "hostsfile": HostsBanEntry
    }

BanEntry = IPTablesCommandBanEntry

def main():
    pass

if __name__ == "__main__":
    main()



#!/usr/bin/env python

import os

class HostsFile(object):

    def __init__(self, path, buffer_writes=0):

        self.filepath = path
        self.hosts_data = None
        self.buffer_writes = buffer_writes

        if buffer_writes:
            self.buffer = []
        else:
            self.buffer = None

    def __load_data(self):
        linedata = []
        with open(self.filepath, "r") as fileobj:
            for line in fileobj.read().split("\n"):
                if line.startswith("#"):
                    continue

                linesplit = line.split()
                if len(linesplit) == 2:
                    # No comment
                    linesplit.append(None)
                if linesplit:
                    linedata.append(linesplit)
        return linedata

    @staticmethod
    def __render_entry(linesplit):
        commentstr = ""

        if linesplit[2] and not linesplit[2].startswith("#"):
            commentstr = "#%s" % linesplit[2]
        elif linesplit[2]:
            commenstr = linesplit[2]

        return "%s %s %s" % (
            linesplit[0],
            linesplit[1],
            commentstr
            )

    def add(self, ipaddress, bantype="ALL", comment=None):
        self.hosts_data = self.__load_data()
        bantype = "%s:" % bantype
        lineentry = self.__render_entry([bantype, ipaddress, comment])

        # if entry is the same, do nothing
        if [bantype, ipaddress, comment] in self.hosts_data:
            return self

        # if entry exists, update/replace.
        for index, entry in enumerate(self.hosts_data):
            if entry[1] == ipaddress:
                self.hosts_data[index][0] = bantype
                self.hosts_data[index][2] = comment
                return self

        if self.buffer_writes:
            self.buffer.append(lineentry)

        # finally, just add it
        if not self.buffer_writes:
            with open(self.filepath, "a") as fileobj:
                fileobj.write(
                    "%s\n" % lineentry
                    )
        elif self.buffer_writes and len(self.buffer) >= self.buffer_writes:
            print "Paging out writes"
            for entry in self.buffer:
                with open(self.filepath, "a") as fileobj:
                    fileobj.write(
                        "%s\n" % entry
                        )

        self.hosts_data += [bantype, ipaddress, comment]

        #TODO: return something better
        return self

    def __add__(self, ipaddress):
        self.hosts_data = self.__load_data()
        comment = None
        bantype = "%s:" % "ALL"
        lineentry = self.__render_entry([bantype, ipaddress, comment])

        # finally, just add it
        with open(self.filepath, "a") as fileobj:
            fileobj.write(
                "%s\n" % lineentry
            )
        self.hosts_data = [bantype, ipaddress, comment]

        #TODO: return something better
        return self

    def __sub__(self, ipaddress):
        out_data = []
        self.hosts_data = self.__load_data()
        for line in self.hosts_data:
            if line[1] != ipaddress:
                out_data.append(line)

        with open(self.filepath, "w") as fileobj:
            for out_line in out_data:
                fileobj.write("%s\n" % self.__render_entry(out_line))

        self.hosts_data = out_data
        #TODO: return something better
        return self

    def __contains__(self, ipaddress):
        self.hosts_data = self.__load_data()
        for line in self.hosts_data:
            if len(line) > 1 and line[1] == ipaddress:
                return True
        return False

    def __len__(self):
        num_lines = 0
        self.hosts_data = self.__load_data()
        for line in self.hosts_data:
            if not line[0].startswith("#"):
                num_lines += 1
        return num_lines

    def __iter__(self):
        self.hosts_data = self.__load_data()
        for line in self.hosts_data:
            yield line

    def __nonzero__(self):
        if self.__len__() > 0:
            return True
        else:
            return False

    #TODO __setitem__
    def __getitem__(self, ipaddress):
        self.hosts_data = self.__load_data()
        for line in self.hosts_data:
            if len(line) > 1 and line[1] == ipaddress:
                return line
        return None

    def __bool__(self):
        return self.__nonzero__()

class HostsDeny(HostsFile):

    def __init__(self, buffer_writes=0):
        HostsFile.__init__(self, "/etc/hosts.deny",
                           buffer_writes=buffer_writes)

class HostsAllow(HostsFile):

    def __init__(self, buffer_writes=0):
        HostsFile.__init__(self, "/etc/hosts.allow",
                           buffer_writes=buffer_writes)

if __name__ == "__main__":
    # this should be a unit test oh well
    a = HostsFile("./hosts.deny")
    sought = "218.108.85.250"
    print len(a)
    if sought in a:
        print "%s is in the file" % sought
    else:
        print "%s is not in the file" % sought

    a -= sought
    if sought in a:
        print "%s is in the file" % sought
    else:
        print "%s is not in the file" % sought

    a += sought
    if sought in a:
        print "%s is in the file" % sought
    else:
        print "%s is not in the file" % sought



import os
from setuptools import setup

setup(
    name = "swabber",
    version = "1.3.4",
    author = "Hugh Nowlan",
    author_email = "nosmo@nosmo.me",
    description = "Pubsub-based host banning interface",
    license = "N(C)",
    keywords = "iptables banning pubsub",
    url = "http://github.com/equalitie/swabber",
    packages=['swabber'],
    install_requires=[
        "python-iptables",
        "pyzmq",
        "ipaddr",
        "pyyaml"
    ],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Topic :: Security",
        "Intended Audience :: System Administrators",
        ],
    scripts = ["swabberd"],
    # Not distributing configuration and init script with the package
    # because data_files is fucking USELESS. Can't rename files, can't
    # address files with a relative path, must have the init script with the
    # same name as the actual script getting installed into bin/. Fuck
    # that, fuck off, fuck computers.
    )
